Reference Selection Keywords for READNOISE
++++++++++++++++++++++++++++++++++++++++++
CRDS selects appropriate READNOISE references based on the following keywords.
READNOISE is not applicable for instruments not in the table.
All keywords used for file selection are *required*.

========== ================================================
Instrument Keywords
========== ================================================
WFI        instrument, detector, date, time
========== ================================================
